import subprocess
from pathlib import Path

import click
from tqdm import tqdm

from kks.util import get_solution_directory, format_file, test_number_to_name, find_test_pairs, get_matching_suffix


@click.command(short_help='Generate tests')
@click.option('-o', '--output-only', is_flag=True,
              help='If specified, only solution will be run. Useful to generate output for manually created tests')
@click.option('-g', '--generator', type=click.Path(exists=True),
              help='Script, used to generate .in files')
@click.option('-s', '--solution', type=click.Path(exists=True),
              help='Script, used to generate .out files')
@click.option('-t', '--test', 'tests', type=int, multiple=True,
              help='Test number to generate')
@click.option('-r', '--range', 'test_range', type=int, nargs=2,
              help='Tests to generate')
@click.option('-f', '--force', is_flag=True,
              help='Overwrite .in files')
@click.argument('gen_args', nargs=-1, type=click.UNPROCESSED)
def gen(output_only, generator, solution, tests, test_range, force, gen_args):
    """
    Generate tests

    \b
    Example usage:
      kks gen -t 1337
      kks gen -r 1 100
      kks gen -t 1 -o
      kks gen -g gen.py -s solve.py -r 1 50 -f
    """

    directory = get_solution_directory()

    generator = Path(generator or directory / 'gen.py')
    solution = Path(solution or directory / 'solve.py')

    if not generator.exists() and not output_only:
        click.secho(f'Generator {generator} does not exist', fg='red', err=True)
        return

    if not solution.exists():
        click.secho(f'Solution {solution} does not exist', fg='red', err=True)
        return

    test_pairs = find_tests_to_gen(directory, tests, test_range, output_only)
    test_pairs = sorted(test_pairs)

    generated_tests = 0

    t = tqdm(test_pairs, leave=False)
    for input_file, output_file in t:
        if not output_only and input_file.exists() and not force:
            t.clear()
            click.secho(f'Input file ' + format_file(input_file), fg='yellow', err=True, nl=False)
            click.secho(f' already exists, skipping. Specify -f to overwrite', fg='yellow', err=True)
            continue

        if output_file and output_file.exists() and not force:
            t.clear()
            click.secho(f'Output file ' + format_file(output_file), fg='yellow', err=True, nl=False)
            click.secho(f' already exists, skipping. Specify -f to overwrite', fg='yellow', err=True)
            continue

        t.set_description(f'Generating test {format_file(input_file)}')

        if not output_only:
            with input_file.open('w') as f:
                args = [input_file.stem] + list(gen_args)
                if run_binary(generator, args, stdout=f) is None:
                    return

        with input_file.open('r') as f_in, output_file.open('w') as f_out:
            args = [input_file.stem]
            if run_binary(solution, args, stdin=f_in, stdout=f_out) is None:
                return

        generated_tests += 1

    click.secho(f'Generated {generated_tests} tests!', fg='green', err=True)


def find_tests_to_gen(directory, tests, test_range, output_only):
    """Находит существующие тесты"""
    tests_dir = directory / 'tests'
    tests_dir.mkdir(exist_ok=True)

    if not tests and not test_range:
        test_range = (1, 100)

    test_numbers = list(tests)
    if test_range:
        l, r = sorted(test_range)
        test_numbers += list(range(l, r + 1))

    test_names = [test_number_to_name(number) for number in test_numbers]

    pairs = list(find_test_pairs(tests_dir, test_names))

    used_names = [input_file.stem for input_file, _ in pairs]
    not_used_names = set(test_names) - set(used_names)

    def output_file_for_input_file(input_file):
        return input_file.with_suffix(get_matching_suffix(input_file.suffix))

    result = []

    # Добавляем выходной файл, если его не существует
    result += [
        (input_file, output_file or output_file_for_input_file(input_file))
        for input_file, output_file in pairs
    ]

    # Если пары совсем не существует, добавляем
    result += [
        ((tests_dir / test_name).with_suffix('.in'),
         (tests_dir / test_name).with_suffix('.out'))
        for test_name in not_used_names
    ]

    return result


def run_binary(binary, args, stdin=None, stdout=None):
    ext = binary.suffix

    interpreter = 'python3' if ext in ['.py', '.py3'] \
        else 'bash' if ext in ['.sh'] \
        else None

    if interpreter is None:
        click.secho(f'Cant run unrecognized script {format_file(binary)}', fg='red', err=True)
        return None

    process = subprocess.run([interpreter, binary] + args, stdin=stdin, stdout=stdout)

    if process.returncode != 0:
        click.secho('Script exited with code ' +
                    click.style(str(process.returncode), fg='red', bold=True) +
                    ' (args: ' + ' '.join(args) + ')', fg='yellow')
        return None

    return process



import subprocess

import click

ARGS = [
    'gcc',
    '-g',
    '-Werror',
    '-Wall',
    '-Wextra',
    '-fsanitize=address',
    '-fsanitize=undefined',
    '-fno-sanitize-recover=all',  # for RE in case of UB
]


def compile_solution(directory):
    c_files = list(directory.glob('*.c'))

    if len(c_files) == 0:
        click.secho('No .c files found', fg='yellow', err=True)
        return None

    click.secho('Compiling... ', fg='green', err=True, nl=False)

    binary = compile_c(directory, c_files)

    if binary is None:
        click.secho('Compilation failed!', fg='red', err=True)
        return None

    binary_name = click.style(binary.relative_to(directory).as_posix(), fg='red', bold=True)
    click.secho(f'Successfully compiled binary {binary_name}', fg='green', err=True)

    return binary


def compile_c(workdir, files):
    filenames = [path.relative_to(workdir) for path in files]

    command = ARGS + filenames
    p = subprocess.run(command, cwd=workdir)

    if p.returncode != 0:
        return None

    return workdir / 'a.out'





import subprocess
from os import environ
from pathlib import Path

import click
from tqdm import tqdm

from kks.binary import compile_solution
from kks.util import get_solution_directory, print_diff, find_test_output, format_file, test_number_to_name, \
    find_test_pairs


@click.command(short_help='Test solutions')
@click.option('-s', '--sample', is_flag=True,
              help='Test only sample')
@click.option('-t', '--test', 'tests', type=int, multiple=True,
              help='Test numbers to run (multiple are allowed)')
@click.option('-r', '--range', 'test_range', type=int, nargs=2,
              help='Tests to run')
@click.option('-f', '--file', 'files', type=click.Path(), multiple=True,
              help='Test files')
@click.option('-c', '--continue', 'cont', is_flag=True,
              help='Continue running after error')
def test(tests, test_range, files, sample, cont):
    """
    Test solution

    \b
    Example usage:
        kks test
        kks test -s
        kks test -t 0 -t 2 -t 3
    """

    files = [Path(f) for f in files]

    directory = get_solution_directory()
    if directory is None:
        return

    binary = compile_solution(directory)
    if binary is None:
        return

    tests = find_tests_to_run(directory, files, tests, test_range, sample)
    if tests is None:
        return

    if not tests:
        click.secho('No tests found!', fg='red')
        return

    successful_count = 0
    ran_count = 0

    test_env = dict(environ, ASAN_OPTIONS="color=always")
    t = tqdm(tests, leave=False)
    for input_file, output_file in t:
        if sample:
            t.clear()
            with input_file.open('r') as f:
                input_data = f.read()
            click.secho("Sample input:", bold=True)
            click.secho(input_data)
            with output_file.open('r') as f:
                output_data = f.read()
            click.secho("Sample output:", bold=True)
            click.secho(output_data)

        t.set_description(f'Running {format_file(input_file)}')

        is_success = run_test(binary, input_file, output_file, test_env)

        ran_count += 1
        successful_count += is_success

        if not cont and not is_success:
            t.close()
            break

    color = 'red' if ran_count != successful_count else 'green'
    click.secho(f'Tests passed: {successful_count}/{ran_count}', fg=color, bold=True)


def run_test(binary, input_file, output_file, env):
    with input_file.open('r') as input_f:
        process = subprocess.run([binary.absolute()], stdin=input_f, capture_output=True, env=env)

    if process.returncode != 0:
        error_output = process.stderr.decode('utf-8')
        click.secho('RE', fg='red', bold=True)
        click.secho(f'Process exited with code {process.returncode}', fg='red')
        if error_output:
            click.secho(error_output)
        return False

    with output_file.open('r') as output_f:
        expected_output = output_f.read()

    actual_output = process.stdout.decode('utf-8')

    if expected_output != actual_output:
        click.secho('WA', fg='red', bold=True)
        print_diff(expected_output, actual_output, 'expected', 'actual')
        click.secho()
        return False

    return True


def find_tests_to_run(directory, files, tests, test_range, sample):
    """
    Возвращает тесты для команды test
    :param directory: Папка задачи
    :param files: Входные файлы тестов
    :param tests: Номера тестов
    :param test_range: Промежуток номеров тестов
    :param sample: Вернуть только семпл
    :return: Пары (input, output)
    """

    result = set()

    need_files = not sample and len(files) != 0
    need_numbers = not sample and tests or test_range

    # need file tests
    if need_files:
        for input_file in files:
            if not input_file.is_file():
                click.secho(f'File {format_file(input_file)} not found', fg='red', err=True)
                return None
            output_file = find_test_output(input_file)
            if output_file is None:
                click.secho(f'No output for test file {format_file(input_file)}', fg='yellow', err=True)
                continue
            result.add((input_file, output_file))

    test_names = None

    # number tests
    if need_numbers:
        test_numbers = list(tests)
        if test_range:
            l, r = sorted(test_range)
            test_numbers += list(range(l, r + 1))

        if sample:
            test_numbers += [0]

        test_names = [test_number_to_name(number) for number in test_numbers]

    if sample:
        test_names = ['000']

    # need numbers or need all
    if need_numbers or not need_files:
        tests_dir = directory / 'tests'
        if not tests_dir.is_dir():
            click.secho(f'Not a directory: {format_file(tests_dir)}', fg='red', err=True)
            return None

        for input_file, output_file in find_test_pairs(tests_dir, test_names):
            if output_file is not None:
                result.add((input_file, output_file))
            else:
                click.secho(f'Test {format_file(input_file)} has no output', fg='yellow', err=True)

    return sorted(result)



import click

from kks.ejudge import ejudge_summary, ejudge_sample
from kks.util import get_valid_session, load_links, find_workspace


@click.command()
def sync():
    """Parse problems from ejudge"""

    workspace = find_workspace()

    if workspace is None:
        click.secho('You have to run sync under kks workspace (use "kks init" to create one)', fg='red', err=True)
        return

    session = get_valid_session()
    if session is None:
        return

    links = load_links()
    if links is None:
        click.secho('Auth data is invalid, use "kks auth" to authorize', fg='red', err=True)
        return

    problems = ejudge_summary(links, session)

    for problem in problems:
        contest, number = problem.short_name.split('-')

        task_dir = workspace / contest / number

        if task_dir.exists():
            if task_dir.is_dir():
                click.secho(f'Directory {task_dir.relative_to(workspace)} already exists, skipping',
                            fg='yellow', err=True)
            else:
                click.secho(f'File {task_dir.relative_to(workspace)} exists, skipping',
                            fg='red', err=True)
            continue

        click.secho('Creating directories for task ' + click.style(problem.name, fg='blue', bold=True), err=True)

        task_dir.mkdir(parents=True, exist_ok=True)

        main = task_dir / f'{problem.short_name}.c'
        main.touch()

        gen = task_dir / 'gen.py'
        gen.touch()

        solve = task_dir / 'solve.py'
        solve.touch()

        tests_dir = task_dir / 'tests'
        tests_dir.mkdir(exist_ok=True)

        input_data, output_data = ejudge_sample(problem.href, session)

        if input_data is not None:
            with (tests_dir / '000.in').open('w') as f:
                f.write(input_data)

        if output_data is not None:
            output_data += '\n'
            with (tests_dir / '000.out').open('w') as f:
                f.write(output_data)

    click.secho('Sync done!', fg='green')



import click

from kks.ejudge import ejudge_standings
from kks.util import get_valid_session, load_links


@click.command()
def top():
    """Parse and display user standings"""

    session = get_valid_session()
    if session is None:
        return

    links = load_links()
    if links is None:
        click.secho('Auth data is invalid, use "kks auth" to authorize', fg='red', err=True)
        return

    standings = ejudge_standings(links, session)

    row_format = "{:>6}  {:25} {:>7} {:>6}  {}"

    click.secho(row_format.format("Place", "User", "Solved", "Score", "Tasks"), fg='white', bold=True)

    for row in standings:
        tasks = ' '.join([
            click.style('{:>3}'.format(task.score or ''), fg=task.color(), bold=task.bold())
            for task in row.tasks
        ])

        string = row_format.format(row.place, row.user, row.solved, row.score, tasks)
        click.secho(string, fg=row.color(), bold=row.bold())





import subprocess

import click

from kks.util import get_solution_directory, get_clang_style_string, print_diff


@click.command(short_help='Lint solution')
@click.option('--diff/--no-diff', is_flag=True, default=True,
              help='Show lint diff')
def lint(diff):
    """
    Lint solution in current task directory using clang-format.

    clang-format has to be present in PATH.

    If file ~/.kks/.clang-format exists, it will be passed to clang-format.
    Otherwise, default config (hardcoded in util.py) is used.
    """

    directory = get_solution_directory()
    if directory is None:
        return

    c_files = list(directory.glob('*.c'))
    if len(c_files) == 0:
        click.secho('No *.c files found', fg='yellow', err=True)
        return

    before = {}
    if diff:
        for file in c_files:
            with file.open('r') as f:
                before[file] = f.read()

    file_names = [file.as_posix() for file in c_files]

    files_string = click.style(' '.join(file_names), fg='blue', bold=True)
    click.secho('Formatting files ' + files_string)

    style_string = '--style=' + get_clang_style_string()
    process = subprocess.run(['clang-format', '-i', style_string] + file_names)

    if process.returncode != 0:
        click.secho(f'Clang-format exited with exit-code {process.returncode}', fg='red', err=True)
        return

    after = {}
    if diff:
        for file in c_files:
            with file.open('r') as f:
                after[file] = f.read()

        for file in c_files:
            print_diff(before[file], after[file], file.as_posix(), file.as_posix())

    click.secho(f'Successfully formatted!', fg='green', err=True)



from setuptools import setup, find_packages

setup(
    name='kokos',
    description='KoKoS helper tool',
    author='Vyacheslav Boben',
    version='1.0.12',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'Click',
        'requests',
        'configparser',
        'colorama',
        'tqdm',
        'beautifulsoup4',
    ],
    entry_points='''
        [console_scripts]
        kks=kks.cli:cli
    ''',
)
